#  (c) Copyright 2015 Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import re
from cinderdiags import constant


logger = logging.getLogger(__name__)


def check_all(client, node, service):
    """Check for default packages on cinder or nova node

    :param client: ssh client
    :param node: node being checked
    :param pkg_info: tuple of ('package name', 'minimum version')
    :return: list of dictionaries
    """
    defaults = {
        'cinder': constant.CINDER_PACKAGES,
        'nova': constant.NOVA_PACKAGES,
    }

    checked = []
    checker = get_check_type(client, node)
    if checker is not None:
        for pkg in defaults[service]:
            check = checker(client, node, pkg)
            if check['installed'] == 'unknown':
                check = pip_check(client, node, pkg)
            checked.append(check)
    else:
        checked.append({
            'node': node,
            'name': 'ERROR',
            'installed': 'ERROR',
            'version': 'ERROR',
        })
    return checked


def check_one(client, node, pkg_info):
    """Check for a single package on a single node

    :param client: ssh client
    :param node: node being checked
    :param pkg_info: tuple of ('package name', 'minimum version')
    :return: dictionary
    """

    checker = get_check_type(client, node)
    if checker is not None:
        check = checker(client, node, pkg_info)
        if check['installed'] == 'unknown':
            check = pip_check(client, node, pkg_info)
    else:
        check = {
            'node': node,
            'name': pkg_info[0],
            'installed': 'ERROR',
            'version': 'ERROR',
        }
    return check


def dpkg_check(client, node, pkg_info):
    """Check for packages installed via apt-get (Debian Linux)

    :param client: ssh client
    :param node: node being checked
    :param pkg_info: (name, version)
    :return: dictionary
    """
    pkg = {
        'node': node,
        'name': pkg_info[0],
        'installed': 'unknown',
        'version': 'N/A',
    }

    names = [x.strip() for x in pkg_info[0].split('||')]
    try:
        for name in names:
            logger.info("Checking for software package '%s' on node %s using "
                        "dpkg-query" % (name, node))
            response = client.execute("dpkg-query -W -f='${Status}"
                                      "${Version}' " + name)
            if 'install ok installed' in response:
                pkg['installed'] = 'pass'
                pkg['name'] = name
                if pkg_info[1]:
                    pattern = re.compile('\D([\d\.]+\d)\D')
                    pkg['version'] = version_check(response,
                                                   pattern,
                                                   pkg_info[1])
                break

    except Exception as e:
        logger.warning("%s -- Unable to check %s on node %s" % (e,
                                                                pkg['name'],
                                                                node))
        pkg['installed'] = 'ERROR'
        pkg['version'] = 'ERROR'
        pass
    return pkg


def yum_check(client, node, pkg_info):
    """Check for packages installed via yum (RedHat Linux)

    :param client: ssh client
    :param node: node being checked
    :param pkg_info: (name, version)
    :return: dictionary
    """
    pkg = {
        'node': node,
        'name': pkg_info[0],
        'installed': 'unknown',
        'version': 'N/A',
    }

    names = [x.strip() for x in pkg_info[0].split('||')]
    try:
        for name in names:
            logger.info("Checking for software package '%s' on node %s using "
                        "yum" % (name, node))
            response = client.execute("yum list installed " +
                                      name)
            if 'Available Packages' in response:
                pkg['installed'] = 'fail'
                pkg['name'] = name
            elif 'Installed Packages' in response:
                pkg['installed'] = 'pass'
                pkg['name'] = name
                if pkg_info[1]:
                    pattern = re.compile(name + '\.[\w+]+\D*([\d\.]+\d)\D')
                    pkg['version'] = version_check(response,
                                                   pattern,
                                                   pkg_info[1])
                break

    except Exception as e:
        logger.warning("%s -- Unable to check %s on node %s" % (e,
                                                                pkg['name'],
                                                                node))
        pkg['installed'] = 'ERROR'
        pkg['version'] = 'ERROR'
        pass
    return pkg


def zypper_check(client, node, pkg_info):
    """Check for packages installed via zypper (SUSE Linux)

    :param client: ssh client
    :param node: node being checked
    :param pkg_info: (name, version)
    :return: dictionary
    """
    pkg = {
        'node': node,
        'name': pkg_info[0],
        'installed': 'unknown',
        'version': 'N/A',
    }

    names = [x.strip() for x in pkg_info[0].split('||')]

    try:
        for name in names:
            logger.info("Checking for software package '%s' on node %s using "
                        "zypper" % (name, node))
            response = client.execute("zypper info " +
                                      name)
            if 'Installed: No' in response:
                pkg['installed'] = 'fail'
                pkg['name'] = name
            elif 'Installed: Yes' in response:
                pkg['installed'] = 'pass'
                pkg['name'] = name
                if pkg_info[1]:
                    pattern = re.compile('Version: \D*([\d\.]+\d)\D')
                    pkg['version'] = version_check(response,
                                                   pattern,
                                                   pkg_info[1])
                break

    except Exception as e:
        logger.warning("%s -- Unable to check %s on node %s" % (e,
                                                                pkg['name'],
                                                                node))
        pkg['installed'] = 'ERROR'
        pkg['version'] = 'ERROR'
        pass
    return pkg


def pip_check(client, node, pkg_info):
    """Check for packages installed via pip (pypi packages)

    :param client: ssh client
    :param node: node being checked
    :param pkg_info: (name, version)
    :return: dictionary
    """
    pkg = {
        'node': node,
        'name': pkg_info[0],
        'installed': 'unknown',
        'version': 'N/A',
    }

    names = [x.strip() for x in pkg_info[0].split('||')]
    try:
        for name in names:
            logger.info("Checking for software package '%s' on node %s using "
                        "pip" % (name, node))
            response = client.execute("pip list | grep " + name)
            if response and re.match(name + '\s', response):
                pkg['installed'] = 'pass'
                pkg['name'] = name
                if pkg_info[1]:
                    pattern = re.compile('\(([\d\.]+)\)')
                    pkg['version'] = version_check(response,
                                                   pattern,
                                                   pkg_info[1])
                break
        if pkg['installed'] == 'unknown':
            pkg['installed'] = "fail"
    except Exception as e:
        logger.warning("%s -- Unable to check %s on node %s" % (e,
                                                                pkg['name'],
                                                                node))
        pkg['installed'] = 'ERROR'
        pkg['version'] = 'ERROR'
    return pkg


def version_check(response, pattern, min_v):
    version = pattern.search(response)
    if version is None:
        return 'unknown'
    elif version.group(1) >= min_v:
        return 'pass'
    else:
        return 'fail'


def get_check_type(client, node):
    """Checks the Linux OS flavor and returns the correct check type

    :param client: ssh client
    :param node: node being checked
    :return: function that expects parameters (client, node, pkg_info)
    """
    os_names = {
        "debian": dpkg_check,
        "fedora": yum_check,
        "suse": zypper_check,
    }
    check_type = None
    response = client.execute('cat /etc/*release | grep ^ID_LIKE')
    for os, check in list(os_names.items()):
        if re.compile(os).search(response):
            logger.info("Detected %s operating system on node %s" % (os, node))
            check_type = check
    if check_type is None:
        logger.error("Unable to determine operating system on %s" % node)
    return check_type



#  (c) Copyright 2015 Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import logging

from cinderdiags import conf_reader
from cliff.lister import Lister


class CheckSoftware(Lister):
    """check for required software (and versions) on Cinder and Nova nodes

    output data:
        Node                node names set by user in cli.conf, names must be
                            unique
                                example: [NODE-NAME]
        Software            software package name
                                defaults: hp3parclient, sysfsutils, sg3-utils
        Installed           installation status of the software package
        Version             software package version meets the minimum
                            requirement
    """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(CheckSoftware, self).get_parser(prog_name)
        parser.formatter_class = argparse.RawTextHelpFormatter
        parser.add_argument('-software',
                            dest='name',
                            nargs='?',
                            metavar='PACKAGE-NAME',
                            default='default',
                            help='may also provide "--package-min-version '
                                 'MINIMUM-VERSION"')
        parser.add_argument('-service',
                            dest='serv',
                            default='default',
                            choices=['cinder', 'nova'],
                            help='defaults to checking all nodes')

        parser.add_argument('-test',
                            dest='test',
                            action='store_true',
                            help=argparse.SUPPRESS)

        parser.add_argument('-conf-file',
                            dest='conf',
                            help='location of cli.conf (defaults to '
                                 '/etc/cinderdiags/cli.conf)')

        args, unknown = parser.parse_known_args()
        if args.name:
            parser.add_argument('--package-min-version',
                                dest='version',
                                nargs='?')
        return parser

    def take_action(self, parsed_args):
        reader = conf_reader.Reader(parsed_args.test, parsed_args.conf)
        result = reader.software_check(parsed_args.name,
                                       parsed_args.serv,
                                       parsed_args.version)

        columns = ('Node', 'Software', 'Installed', 'Version')
        data = []
        for pkg in result:
            data.append((
                pkg['node'],
                pkg['name'],
                pkg['installed'],
                pkg['version']),
            )

        return (columns, data)



#  (c) Copyright 2015 Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

DIRECTORY = '/tmp/'
TEST_CLI_CONFIG = '/tmp/cli.conf'
CLI_CONFIG = '/etc/cinderdiags/cli.conf'
NOVA_PACKAGES = [
    ('sysfsutils',  '2.1'),
    ('sg3-utils || sg3_utils', '1.3'),
]
CINDER_PACKAGES = [('hp3parclient', '3.1.1')]
HP3PAR_DRIVERS = [
    'hp_3par_iscsi.HP3PARISCSIDriver',
    'hp_3par_fc.HP3PARFCDriver'
]



#  (c) Copyright 2015 Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from cliff.app import App
from cliff.commandmanager import CommandManager


class CinderDiags(App):
    App.DEFAULT_VERBOSE_LEVEL = 0

    def __init__(self):
        super(CinderDiags, self).__init__(
            description='cinder diagnostics CLI',
            version='0.1',
            command_manager=CommandManager('cliff.cinderdiags'),
            deferred_help=True,
            )

    def initialize_app(self, argv):
        self.LOG.debug('initialize_app')

    def prepare_to_run_command(self, cmd):
        self.LOG.debug('prepare_to_run_command %s', cmd.__class__.__name__)

    def clean_up(self, cmd, result, err):
        self.LOG.debug('clean_up %s', cmd.__class__.__name__)
        if err:
            self.LOG.debug('got an error: %s', err)


def main(argv=sys.argv[1:]):
    myapp = CinderDiags()
    return myapp.run(argv)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#  (c) Copyright 2015 Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
This fakes 3PAR REST API Client behavior to allow for testing of the Cinder
Diagnostics tool without access to a 3PAR array
"""

from hp3parclient import exceptions


class HP3ParClient(object):

    PORT_MODE_TARGET = 2
    PORT_PROTO_ISCSI = 2
    PORT_STATE_READY = 4

    API_URL = 'http://test.ws.url:8080/api/v1'
    USERNAME = 'testuser'
    PASSWORD = 'testpass'
    CPG = 'testCPG'

    def __init__(self, api_url):
        if api_url == self.API_URL:
            pass
        else:
            raise exceptions.HTTPBadRequest('Error communicating with the '
                                            '3PAR WS')

    def login(self, username, password):
        if username == self.USERNAME and password == self.PASSWORD:
            pass
        else:
            raise exceptions.HTTPUnauthorized('invalid username or password')

    def logout(self):
        pass

    def getCPG(self, name):
        if name == self.CPG:
            pass
        else:
            raise exceptions.HTTPNotFound('Invalid input received: CPG')

    def getPorts(self):
        first = {'mode': self.PORT_MODE_TARGET,
                 'linkState': self.PORT_STATE_READY,
                 'protocol': self.PORT_PROTO_ISCSI,
                 'IPAddr': '1.1.1.1'
                 }
        second = {'mode': self.PORT_MODE_TARGET,
                  'linkState': self.PORT_STATE_READY,
                  'protocol': self.PORT_PROTO_ISCSI,
                  'IPAddr': '2.2.2.2'
                  }
        ports = {'members': [first, second]}
        return ports



#  (c) Copyright 2015 Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import logging

from cinderdiags import conf_reader
from cliff.lister import Lister


class CheckOptions(Lister):
    """check 3PAR options in cinder.conf against 3PAR array(s)

    output data:
        Node                node names set by user in cli.conf, must be unique
                                example: [NODE-NAME]
        Backend Section     backend section names set by user in cinder.conf,
                            must be unique per node
                                example: [BACKEND-SECTION-NAME]
        WS API              web service API url for array
                                option: hp3par_api_url
        Credentials         username and password for array
                                option: hp3par_username, hp3par_password
        CPG(s)              CPGs to use for volume creation
                                option: hp3par_cpg
        iSCSI IP(s)         array's iSCSI IP addresses to use
                                option: hp3par_iscsi_ips
        Driver              array's volume driver
                                option: volume_driver
    """

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(CheckOptions, self).get_parser(prog_name)
        parser.formatter_class = argparse.RawTextHelpFormatter
        parser.add_argument('-test',
                            dest='test',
                            action='store_true',
                            help=argparse.SUPPRESS)
        parser.add_argument('-backend-section',
                            dest='name',
                            nargs='?',
                            default='arrays',
                            help='defaults to checking all backend sections')
        parser.add_argument('-conf-file',
                            dest='conf',
                            help='location of cli.conf (defaults to '
                                 '/etc/cinderdiags/cli.conf)')
        return parser

    def take_action(self, parsed_args):
        reader = conf_reader.Reader(parsed_args.test, parsed_args.conf)
        result = reader.options_check(parsed_args.name)
        if len(result) < 1:
            raise ValueError("%s not found" % parsed_args.name)
        columns = (
            'Node',
            'Backend Section',
            'WS API',
            'Credentials',
            'CPG',
            'iSCSI IP(s)',
            'Driver',
        )

        data = []
        for arr in result:
            data.append((
                arr['node'],
                arr['name'],
                arr['url'],
                arr['credentials'],
                arr['cpg'],
                arr['iscsi'],
                arr['driver'],
            ))
        return (columns, data)



#  (c) Copyright 2015 Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import os

from cinderdiags import constant
from cinderdiags import pkg_checks
from cinderdiags import ssh_client
from cinderdiags import hp3par_wsapi_checks as wsapi_checks

from six.moves import configparser


logger = logging.getLogger(__name__)
parser = configparser.ConfigParser()


class Reader(object):

    def __init__(self, is_test=False, path=None):
        self.is_test = is_test
        self.cinder_nodes = []
        self.nova_nodes = []

        if self.is_test:
            path = constant.TEST_CLI_CONFIG
        elif path is None:
            path = constant.CLI_CONFIG
        if os.path.isfile(path):
            parser.read(path)
            self.get_nodes()
            if len(self.cinder_nodes) < 1:
                logger.warning("No Cinder nodes are configured in cli.conf")
            if len(self.nova_nodes) < 1:
                logger.warning("No Nova nodes are configured in cli.conf")
        else:
            raise IOError("%s not found" % constant.CLI_CONFIG)

    def get_nodes(self):
        """Create lists of cinder and nova nodes
        """
        for section_name in list(parser.sections()):
            if parser.get(section_name, 'service').lower() == 'cinder':
                self.cinder_nodes.append(section_name)
            elif parser.get(section_name, 'service').lower() == 'nova':
                self.nova_nodes.append(section_name)

    def get_clients(self, nodes):
        """Create SSH client connections for nodes.
        """
        clients = {}
        for node in nodes:
            try:
                client = ssh_client.Client(parser.get(node, 'host_ip'),
                                           parser.get(node, 'ssh_user'),
                                           parser.get(node, 'ssh_password'))
                clients[node] = client
            except Exception as e:
                logger.warning("%s: %s" % (e, node))
                pass
        return clients

    def copy_files(self, clients):
        """Copy the cinder.conf file of each cinder node to a local directory.

        Location of cinder.conf file is set per node in cli.conf
        """
        files = {}
        for node in self.cinder_nodes:
            try:
                conf_file = clients[node].get_file(parser.get(
                    node, 'conf_source'), constant.DIRECTORY + node)
                files[node] = conf_file
            except Exception as e:
                logger.warning("%s: %s" % (e, node))
        return files

    def software_check(self, name='default', service='default', version=None):
        """Check nodes for installed software packages

        :param name: Name of a software package to check for
        :param service: cinder or nova
        :param version: minimum version of software package
        :return: list of dictionaries
        """
        if service == 'nova':
            checklist = self.nova_nodes
        elif service == 'cinder':
            checklist = self.cinder_nodes
        else:
            checklist = set(self.nova_nodes + self.cinder_nodes)
        clients = self.get_clients(checklist)

        checks = []
        for node in checklist:
            try:
                if name == 'default':
                    checks += pkg_checks.check_all(clients[node],
                                                   node,
                                                   parser.get(node, 'service'))
                else:
                    checks.append(pkg_checks.check_one(clients[node],
                                                       node,
                                                       (name, version)))
            except Exception as e:
                logger.warning("%s: %s" % (e, node))
        self.cleanup(clients)
        return checks

    def options_check(self, section_name='arrays'):
        """Check WS API options in each cinder.conf file

        :param section_name: section name in the cinder.conf file.  Checks
        all by default
        :return: list of dictionaries
        """

        clients = self.get_clients(self.cinder_nodes)
        files = self.copy_files(clients)
        checks = []
        for node in files:
            checker = wsapi_checks.WSChecker(clients[node],
                                             files[node],
                                             node,
                                             self.is_test)
            if section_name == 'arrays':
                checks += checker.check_all()
            else:
                found = checker.check_section(section_name)
                if found:
                    checks.append(found)
        self.cleanup(clients, files)
        return checks

    def cleanup(self, clients, files={}):
        """Delete all copied cinder.conf files and close all SSH connections.
        """
        for node in clients:
            clients[node].disconnect()
        for node in files:
            os.remove(files[node])



#  (c) Copyright 2015 Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import paramiko
import socket


class Client(object):

    def __init__(self, hostName, sshUserName, sshPassword):
        """ Connect and perform action to remote machine using SSH
        """
        try:
            # Connect to remote host
            self.client = paramiko.SSHClient()
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.client.connect(hostName,
                                username=sshUserName,
                                password=sshPassword,
                                timeout=20)

        except socket.error:
            raise Exception("SSH Error: Unable to connect")
        except paramiko.ssh_exception.AuthenticationException:
            raise Exception("SSH Error: Invalid SSH credentials")

    def get_file(self, fromLocation, toLocation):
        """ perform copy action to remote machine using SSH
        """
        if self.client.get_transport() and \
                self.client.get_transport().is_authenticated():
            try:
                # Setup sftp connection and transmit this script
                sftp = self.client.open_sftp()
                sftp.get(fromLocation, toLocation)
                sftp.close()
                return toLocation

            except (IOError, paramiko.ssh_exception.SSHException):
                raise Exception("SSH Error: Unable to copy %s" % fromLocation)

    def disconnect(self):
        """ perform copy action to remote machine using SSH
        """
        self.client.close()

    def execute(self, command):
        # Run the transmitted script remotely without args and show its output.
        # SSHClient.exec_command() returns the tuple (stdin, stdout, stderr)
        if self.client.get_transport() and \
                self.client.get_transport().is_authenticated():
            try:
                resp = self.client.exec_command(command, timeout=20)
                stdout = resp[1].readlines()
                stderr = resp[2].readlines()
                return ''.join(stdout) + ''.join(stderr)

            except (paramiko.ssh_exception.SSHException, socket.timeout):
                raise Exception("SSH Error: Unable to execute remote command "
                                "(%s)" % command)



#  (c) Copyright 2015 Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
This checks the cinder.conf file to find errors in the configuration of 3PAR
arrays.

Requires the python hp3parclient: sudo pip install hp3parclient
Assumes the volume_driver is correctly set
"""

import logging
from six.moves import configparser
from cinderdiags import constant
from cinderdiags import hp3par_testclient as testclient

logger = logging.getLogger(__name__)

try:
    from hp3parclient import client as hpclient
    from hp3parclient import exceptions as hpexceptions
except ImportError:
    logger.error('hp3parclient package not found (pip install hp3parclient)')


class WSChecker(object):

    def __init__(self, client, conf, node, test=False):
        """Tests web service api configurations in the cinder.conf file

        :param conf: location of cinder.conf
        :param node: cinder node the cinder.conf was copied from
        :param test: use testing client
        """
        self.conf = conf
        self.ssh_client = client
        self.node = node
        self.is_test = test
        self.parser = configparser.ConfigParser()
        self.parser.read(self.conf)
        self.hp3pars = []
        for section in self.parser.sections():
            if self.parser.has_option(section, 'volume_driver') \
                    and 'hp_3par' in self.parser.get(section, 'volume_driver'):
                self.hp3pars.append(section)

    def check_all(self):
        """Tests configuration settings for all 3par arrays

        :return: a list of dictionaries
        """
        all_tests = []
        for section_name in self.hp3pars:
            all_tests.append(self.check_section(section_name))
        return all_tests

    def check_section(self, section_name):
        """Runs all WS configuration tests for a section

        :param section_name: from cinder.conf as [SECTION_NAME]
        :return: a dictionary - each property is pass/fail/unknown
        """
        tests = {
            "name": section_name,
            "url": "unknown",
            "credentials": "unknown",
            "cpg": "unknown",
            "iscsi": "unknown",
            "node": self.node,
            "driver": "unknown"
        }
        if section_name in self.hp3pars:
            logger.info("Checking 3PAR options for node '%s' backend section "
                         "%s" % (self.node, section_name))
            tests["driver"] = self.has_driver(section_name)
            client = self.get_client(section_name, self.is_test)
            if client:
                tests["url"] = "pass"
                if self.cred_is_valid(section_name, client):
                    tests["credentials"] = "pass"
                    tests["cpg"] = self.cpg_is_valid(section_name, client)
                    if 'iscsi' in self.parser.get(section_name,
                                                  'volume_driver'):
                        tests["iscsi"] = self.iscsi_is_valid(section_name,
                                                             client)
                    client.logout()
                else:
                    tests["credentials"] = "fail"
            else:
                tests["url"] = "fail"
            if 'hp_3par_fc' in self.parser.get(section_name, 'volume_driver'):
                tests["iscsi"] = "N/A"
            return tests
        else:
            return None

# Config testing methods check if option values are valid
    def get_client(self, section_name, test):
        """Tries to create a client and verifies the api url

        :return: The client if url is valid, None if invalid/missing
        """
        try:
            url = self.parser.get(section_name, 'hp3par_api_url')
            logger.info("Attempting to connect to %s..." % url)
            if test:
                cl = testclient.HP3ParClient(url)
            else:
                cl = hpclient.HP3ParClient(url)
            return cl
        except (hpexceptions.UnsupportedVersion, hpexceptions.HTTPBadRequest,
                TypeError) as e:
            logger.info("Failed to connect to hp3par_api_url for node '%s' "
                        "backend section '%s' --- %s" % (self.node,
                                                         section_name, e))
            return None
        except configparser.NoOptionError:
            logger.info("No hp3par_api_url provided for node '%s' backend "
                        "section '%s'" % (self.node, section_name))
            return None

    def cred_is_valid(self, section_name, client):
        """Tries to login to the client to verify credentials

        :return: True if credentials are valid, False if invalid/missing
        """
        try:
            client.login(self.parser.get(section_name, 'hp3par_username'),
                         self.parser.get(section_name, 'hp3par_password'))
            return True
        except (hpexceptions.HTTPForbidden,
                hpexceptions.HTTPUnauthorized):
            logger.info("Incorrect hp3par_username or hp3par_password "
                        "provided for node '%s' backend section '%s'" %
                        (self.node, section_name))
            return False
        except configparser.NoOptionError:
            logger.info("No hp3par_username or hp3par_password provided for "
                        "node '%s' backend section '%s'" % (self.node,
                                                            section_name))
            return False

    def cpg_is_valid(self, section_name, client):
        """Tests to see if a cpg exists on the 3PAR array to verify cpg name

        :return: string
        """
        result = "pass"
        try:
            cpg_list = [x.strip() for x in
                        self.parser.get(section_name, 'hp3par_cpg').split(',')]
            logger.info("Checking hp3par_cpg option for node '%s' backend "
                         "section '%s'" % (self.node, section_name))
            for cpg in cpg_list:
                try:
                    client.getCPG(cpg)
                except hpexceptions.HTTPNotFound:
                    logger.info("Node '%s' backend section '%s' hp3par_cpg "
                                "contains an invalid CPG name: '%s'" %
                                (self.node, section_name, cpg))
                    result = "fail"
        except configparser.NoOptionError:
            logger.info("No hp3par_cpg provided for node '%s' backend section "
                        "'%s'" %
                        (self.node, section_name))
            result = "fail"
        return result

    def iscsi_is_valid(self, section_name, client):
        """Gets the iSCSI target ports from the client, checks the iSCSI IPs.

        :return: string
        """
        valid_ips = []
        result = "pass"
        try:
            ip_list = [x.strip() for x in
                       self.parser.get(section_name,
                                       'hp3par_iscsi_ips').split(',')]
            logger.info("Checking iSCSI IP addresses for node '%s' backend "
                         "section '%s'" % (self.node, section_name))
            for port in client.getPorts()['members']:
                if (port['mode'] == client.PORT_MODE_TARGET and
                        port['linkState'] == client.PORT_STATE_READY and
                        port['protocol'] == client.PORT_PROTO_ISCSI):
                    valid_ips.append(port['IPAddr'])
            for ip_addr in ip_list:
                ip = ip_addr.split(':')
                if ip[0] not in valid_ips:
                    logger.info("Node '%s' backend section '%s' "
                                "hp3par_iscsi_ips contains an invalid iSCSI "
                                "IP '%s'" % (self.node, section_name, ip))
                    result = "fail"
        except configparser.NoOptionError:
            logger.info("No hp3par_iscsi_ips provided for node '%s' backend "
                        "section '%s" % (self.node, section_name))
            result = "fail"
        return result

    def has_driver(self, section_name):
        """Checks that the volume_driver is installed

        :return: string
        """
        try:
            volume_driver = self.parser.get(section_name, 'volume_driver')
            path = volume_driver.split('.')
            if ("%s.%s" % (path[-2], path.pop())) in constant.HP3PAR_DRIVERS:
                path = '/'.join(path)
                logger.info("Checking for driver at '%s' for node '%s' "
                             "backend section '%s'" % (self.node, path,
                                                       section_name))
                response = self.ssh_client.execute('locate ' + path)

                if path in response:
                    result = "pass"
                elif ('command not found' or 'command-not-found') in response:
                    logger.warning("Could not check '%s' driver on node'%s'. "
                                   "Make sure that 'mlocate' is installed on "
                                   "the node." % (section_name, self.node))
                    result = "unknown"
                else:
                    logger.info("Node '%s' backend section '%s' volume_driver "
                                "contains an "
                                "invalid driver location: '%s'" % (
                        self.node, section_name,  path))
                    result = "fail"
            else:
                logger.info("Node '%s' backend section '%s' volume_driver is "
                            "invalid " % (self.node, section_name))
                result = "fail"

        except configparser.NoOptionError:
            logger.info("No volume_diver provided for node '%s' backend "
                        "section '%s'" % (self.node, section_name))
        return result
